# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 The Python Software Foundation.
# See LICENSE.txt and CONTRIBUTORS.txt.
#
import sys

import unittest

_ver = sys.version_info[:2]

if _ver[0] < 3:
    import Queue as queue
    from SimpleXMLRPCServer import SimpleXMLRPCServer
    from SimpleHTTPServer import SimpleHTTPRequestHandler
    from BaseHTTPServer import HTTPServer
    text_type = unicode
    from urllib import unquote
    from urllib2 import Request
    from urlparse import urlparse
else:
    import queue
    from xmlrpc.server import SimpleXMLRPCServer
    from http.server import HTTPServer, SimpleHTTPRequestHandler
    text_type = str
    from urllib.parse import urlparse, unquote
    from urllib.request import Request



# Copyright (C) 2024 Stewart Miles
# Licensed to the Python Software Foundation under a contributor agreement.
# See LICENSE.txt and CONTRIBUTORS.txt.
import codecs
import os
import json
from setuptools import Extension, setup
from setuptools.command import egg_info
import sys


EMBED_EXTENSIONS_METADATA = (
    int(os.getenv('MINIMEXT_EMBED_EXTENSIONS_METADATA', '0')))


class EggInfo(egg_info.egg_info):
    """egg_info command that optionally writes extensions metadata.

    distlib.wheel.Wheel attempts to read the list of extensions from the
    undocumented JSON EXTENSIONS metadata file.

    This command will add the special file JSON EXTENSIONS metadata file to the
    *.dist-info directory in the wheel if the
    MINIMEXT_EMBED_EXTENSIONS_METADATA environment variable is set to 1.
    """

    def run(self):
        egg_info.egg_info.run(self)
        if EMBED_EXTENSIONS_METADATA:
            build_ext = self.get_finalized_command('build_ext')
            extensions_dict = {
                ext_module.name: build_ext.get_ext_filename(ext_module.name)
                for ext_module in self.distribution.ext_modules
            }
            with open(os.path.join(self.egg_info, 'EXTENSIONS'), 'wb') as (
                    extensions_file):
                json.dump(extensions_dict,
                          codecs.getwriter('utf-8')(extensions_file),
                          indent=2)


setup(
    name='minimext' + ('_metadata' if EMBED_EXTENSIONS_METADATA else ''),
    version='0.1',
    description='Calculates Fibonacci numbers.',
    long_description=(
        'Distribution that provides calculate.fib() and calculate_py.fib() '
        'which calculate Fibonacci numbers. minimext.calculate is implemented '
        'as a C extension to test distlib.wheel.Wheel.mount().'),
    packages=['minimext'],
    ext_modules=[
        Extension(name='minimext.calculate',
                  sources=['calculate.c'],
                  py_limited_api=True,
                  define_macros=[
                      ('Py_LIMITED_API', str(sys.version_info.major)),
                  ]),
    ],
    # The extension uses the limited API so tag the wheel as compatible with
    # Python 3.2 and later.
    #
    # Unfortunately the py_limited_api argument to Extension does not mark the
    # wheel as supporting the limited API, so set the see compatibility
    # manually.
    options={'bdist_wheel': {'py_limited_api': 'cp32'}},
    cmdclass={'egg_info': EggInfo},
)



# -*- coding: utf-8 -*-

class Towel(object):
    """A towel, that one should never be without."""

    def __init__(self, color='tie-dye'):
        self.color = color
        self.wrapped_obj = None

    def wrap(self, obj):
        """Wrap an object up in our towel."""
        self.wrapped_obj = obj

    def unwrap(self):
        """Unwrap whatever is in our towel and return whatever it is."""
        obj = self.wrapped_obj
        self.wrapped_obj = None
        return obj



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Vinay Sajip.
# Licensed to the Python Software Foundation under a contributor agreement.
# See LICENSE.txt and CONTRIBUTORS.txt.
#
from __future__ import unicode_literals

import io
import os
import re
import shutil
import subprocess
import sys
import tempfile

from compat import unittest
from support import DistlibTestCase

from distlib import __version__, DistlibException
from distlib.compat import ZipFile, sysconfig, fsencode
from distlib.database import DistributionPath
from distlib.manifest import Manifest
from distlib.metadata import Metadata, METADATA_FILENAME, LEGACY_METADATA_FILENAME
from distlib.scripts import ScriptMaker
from distlib.util import get_executable
from distlib.wheel import (Wheel, PYVER, IMPVER, ARCH, ABI, COMPATIBLE_TAGS, IMP_PREFIX, VER_SUFFIX,
                           is_compatible, _get_glibc_version)

try:
    with open(os.devnull, 'wb') as junk:
        subprocess.check_call(['pip', '--version'], stdout=junk, stderr=subprocess.STDOUT)
    PIP_AVAILABLE = True
except Exception:
    PIP_AVAILABLE = False

HERE = os.path.abspath(os.path.dirname(__file__))

EGG_INFO_RE = re.compile(r'(-py\d\.\d)?\.egg-info', re.I)


def pip_version():
    result = None
    fd, fn = tempfile.mkstemp(prefix='distlib-test-', suffix='.txt')
    try:
        os.close(fd)
        with open(fn, 'wb') as out:
            subprocess.check_call(['pip', '--version'], stdout=out, stderr=subprocess.STDOUT)
        with io.open(fn, encoding='utf-8') as f:
            data = f.read().split()
        assert data[0] == 'pip'
        parts = data[1].split('.')
        result = []
        for p in parts:
            if p.isdigit():
                result.append(int(p))
            else:
                result.append(p)
        result = tuple(result)
    finally:
        os.remove(fn)
    return result


def convert_egg_info(libdir, prefix):
    files = os.listdir(libdir)
    ei = list(filter(lambda d: d.endswith('.egg-info'), files))[0]
    olddn = os.path.join(libdir, ei)
    di = EGG_INFO_RE.sub('.dist-info', ei)
    newdn = os.path.join(libdir, di)
    os.rename(olddn, newdn)
    files = os.listdir(newdn)
    for oldfn in files:
        pn = os.path.join(newdn, oldfn)
        if oldfn == 'PKG-INFO':
            md = Metadata(path=pn)
            mn = os.path.join(newdn, METADATA_FILENAME)
            md.write(mn)
        os.remove(pn)
    manifest = Manifest(os.path.dirname(libdir))
    manifest.findall()
    dp = DistributionPath([libdir])
    dist = next(dp.get_distributions())
    dist.write_installed_files(manifest.allfiles, prefix)


def install_dist(distname, workdir):
    pfx = '--install-option='
    purelib = pfx + '--install-purelib=%s/purelib' % workdir
    platlib = pfx + '--install-platlib=%s/platlib' % workdir
    headers = pfx + '--install-headers=%s/headers' % workdir
    scripts = pfx + '--install-scripts=%s/scripts' % workdir
    data = pfx + '--install-data=%s/data' % workdir
    cmd = [
        'pip', 'install', '--index-url', 'https://pypi.org/simple/', '--timeout', '3', '--default-timeout', '3',
        purelib, platlib, headers, scripts, data, distname
    ]
    result = {
        'scripts': os.path.join(workdir, 'scripts'),
        'headers': os.path.join(workdir, 'headers'),
        'data': os.path.join(workdir, 'data'),
    }
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout, _ = p.communicate()
    if p.returncode:
        raise ValueError('pip failed to install %s:\n%s' % (distname, stdout))
    for dn in ('purelib', 'platlib'):
        libdir = os.path.join(workdir, dn)
        if os.path.isdir(libdir):
            result[dn] = libdir
            break
    convert_egg_info(libdir, workdir)
    dp = DistributionPath([libdir])
    dist = next(dp.get_distributions())
    md = dist.metadata
    result['name'] = md.name
    result['version'] = md.version
    return result


class WheelTestCase(DistlibTestCase):

    def test_valid_filename(self):
        attrs = ('name', 'version', 'buildver', 'pyver', 'abi', 'arch')
        cases = (
            ('pkg-1.0.0-cp32.cp33-noabi-noarch.whl', ('pkg', '1.0.0', '', ['cp32', 'cp33'], ['noabi'], ['noarch'])),
            ('package-1.0.0-cp33-noabi-linux_x86_64.whl', ('package', '1.0.0', '', ['cp33'], ['noabi'],
                                                           ['linux_x86_64'])),
            ('test-1.0-1st-py2.py3-none-win32.whl', ('test', '1.0', '1st', ['py2', 'py3'], ['none'], ['win32'])),
            ('Pillow-2.8.1-cp27-none-macosx_10_6_intel.'
             'macosx_10_9_intel.macosx_10_9_x86_64.macosx_10_10_intel.'
             'macosx_10_10_x86_64.whl', ('Pillow', '2.8.1', '', ['cp27'], ['none'], [
                 'macosx_10_6_intel', 'macosx_10_9_intel', 'macosx_10_9_x86_64', 'macosx_10_10_intel',
                 'macosx_10_10_x86_64'
             ])),
        )

        for name, values in cases:
            w = Wheel(name)
            self.assertEqual(w.wheel_version, (1, 1))
            self.assertEqual(w.filename, name)
            for attr, value in zip(attrs, values):
                self.assertEqual(getattr(w, attr), value)

    def test_invalid_filename(self):
        names = (
            '',
            'package.whl',
            'package-1.0.0-cp32.cp33.whl',
            'package-1.0.0-cp32.cp33.whl',
            'package-1.0.0-cp32.cp33-noabi.whl',
            'package-1.0.0-cp32.cp33-noabi-noarch.zip',
        )

        for name in names:
            self.assertRaises(DistlibException, Wheel, name)

    def test_valid_name(self):
        attrs = ('name', 'version', 'buildver', 'pyver', 'abi', 'arch')
        pyver = PYVER
        cases = (
            ('pkg-1.0.0', ('pkg', '1.0.0', '', [pyver], ['none'], ['any'])),
            ('test-1.0-1st', ('test', '1.0', '1st', [pyver], ['none'], ['any'])),
            (None, ('dummy', '0.1', '', [pyver], ['none'], ['any'])),
        )

        ENDING = '-%s-none-any.whl' % PYVER
        for name, values in cases:
            w = Wheel(name)
            self.assertEqual(w.wheel_version, (1, 1))
            self.assertTrue(w.filename.endswith(ENDING))
            for attr, value in zip(attrs, values):
                self.assertEqual(getattr(w, attr), value)

    def test_compatible_tags(self):
        self.assertEqual(PYVER, 'py%d%d' % sys.version_info[:2])
        tags = COMPATIBLE_TAGS
        self.assertIn((PYVER, 'none', 'any'), tags)
        self.assertIn((PYVER[:-1], 'none', 'any'), tags)
        this_arch = filter(lambda o: o[-1] == ARCH, tags)
        self.assertTrue(this_arch)
        if sys.platform.startswith('linux'):
            arch = ARCH.replace('linux_', '')
            parts = _get_glibc_version()
            if len(parts) == 2:
                self.assertTrue(filter(lambda o: o[-1] == 'manylinux_%s_%s_%s' % (parts[0], parts[1], arch), tags))
                if parts >= (2, 17):
                    self.assertTrue(filter(lambda o: o[-1] == 'manylinux2014_%s' % arch, tags))
                if parts >= (2, 12):
                    self.assertTrue(filter(lambda o: o[-1] == 'manylinux2010_%s' % arch, tags))
                if parts >= (2, 5):
                    self.assertTrue(filter(lambda o: o[-1] == 'manylinux1_%s' % arch, tags))

    def test_is_compatible(self):
        fn = os.path.join(HERE, 'dummy-0.1-py27-none-any.whl')
        if PYVER in ('py27', 'py30', 'py31'):
            self.assertTrue(is_compatible(fn))
            self.assertTrue(Wheel(fn).is_compatible())
        # use actual wheel names from PyPI.
        wheel_names = [
            'simplejson-3.17.2-cp27-cp27m-macosx_10_13_x86_64.whl',
            'simplejson-3.17.2-cp27-cp27m-manylinux1_i686.whl',
            'simplejson-3.17.2-cp27-cp27m-manylinux1_x86_64.whl',
            'simplejson-3.17.2-cp27-cp27m-manylinux2010_i686.whl',
            'simplejson-3.17.2-cp27-cp27m-manylinux2010_x86_64.whl',
            'simplejson-3.17.2-cp27-cp27mu-manylinux1_i686.whl',
            'simplejson-3.17.2-cp27-cp27mu-manylinux1_x86_64.whl',
            'simplejson-3.17.2-cp27-cp27mu-manylinux2010_i686.whl',
            'simplejson-3.17.2-cp27-cp27mu-manylinux2010_x86_64.whl',
            'simplejson-3.17.2-cp27-cp27m-win32.whl',
            'simplejson-3.17.2-cp27-cp27m-win_amd64.whl',
            'simplejson-3.17.2-cp33-cp33m-win32.whl',
            'simplejson-3.17.2-cp33-cp33m-win_amd64.whl',
            'simplejson-3.17.2-cp34-cp34m-win32.whl',
            'simplejson-3.17.2-cp34-cp34m-win_amd64.whl',
            'simplejson-3.17.2-cp35-cp35m-manylinux1_i686.whl',
            'simplejson-3.17.2-cp35-cp35m-manylinux1_x86_64.whl',
            'simplejson-3.17.2-cp35-cp35m-manylinux2010_i686.whl',
            'simplejson-3.17.2-cp35-cp35m-manylinux2010_x86_64.whl',
            'simplejson-3.17.2-cp35-cp35m-manylinux2014_aarch64.whl',
            'simplejson-3.17.2-cp35-cp35m-win32.whl',
            'simplejson-3.17.2-cp35-cp35m-win_amd64.whl',
            'simplejson-3.17.2-cp36-cp36m-macosx_10_13_x86_64.whl',
            'simplejson-3.17.2-cp36-cp36m-manylinux1_i686.whl',
            'simplejson-3.17.2-cp36-cp36m-manylinux1_x86_64.whl',
            'simplejson-3.17.2-cp36-cp36m-manylinux2010_i686.whl',
            'simplejson-3.17.2-cp36-cp36m-manylinux2010_x86_64.whl',
            'simplejson-3.17.2-cp36-cp36m-manylinux2014_aarch64.whl',
            'simplejson-3.17.2-cp36-cp36m-win32.whl',
            'simplejson-3.17.2-cp36-cp36m-win_amd64.whl',
            'simplejson-3.17.2-cp37-cp37m-macosx_10_14_x86_64.whl',
            'simplejson-3.17.2-cp37-cp37m-manylinux1_i686.whl',
            'simplejson-3.17.2-cp37-cp37m-manylinux1_x86_64.whl',
            'simplejson-3.17.2-cp37-cp37m-manylinux2010_i686.whl',
            'simplejson-3.17.2-cp37-cp37m-manylinux2010_x86_64.whl',
            'simplejson-3.17.2-cp37-cp37m-manylinux2014_aarch64.whl',
            'simplejson-3.17.2-cp37-cp37m-win32.whl',
            'simplejson-3.17.2-cp37-cp37m-win_amd64.whl',
            'simplejson-3.17.2-cp38-cp38-macosx_10_14_x86_64.whl',
            'simplejson-3.17.2-cp38-cp38-manylinux1_i686.whl',
            'simplejson-3.17.2-cp38-cp38-manylinux1_x86_64.whl',
            'simplejson-3.17.2-cp38-cp38-manylinux2010_i686.whl',
            'simplejson-3.17.2-cp38-cp38-manylinux2010_x86_64.whl',
            'simplejson-3.17.2-cp38-cp38-manylinux2014_aarch64.whl',
            'Pillow-7.2.0-cp35-cp35m-macosx_10_10_intel.whl',
            'Pillow-7.2.0-cp35-cp35m-manylinux1_i686.whl',
            'Pillow-7.2.0-cp35-cp35m-manylinux1_x86_64.whl',
            'Pillow-7.2.0-cp35-cp35m-manylinux2014_aarch64.whl',
            'Pillow-7.2.0-cp35-cp35m-win32.whl',
            'Pillow-7.2.0-cp35-cp35m-win_amd64.whl',
            'Pillow-7.2.0-cp36-cp36m-macosx_10_10_x86_64.whl',
            'Pillow-7.2.0-cp36-cp36m-manylinux1_i686.whl',
            'Pillow-7.2.0-cp36-cp36m-manylinux1_x86_64.whl',
            'Pillow-7.2.0-cp36-cp36m-manylinux2014_aarch64.whl',
            'Pillow-7.2.0-cp36-cp36m-win32.whl',
            'Pillow-7.2.0-cp36-cp36m-win_amd64.whl',
            'Pillow-7.2.0-cp37-cp37m-macosx_10_10_x86_64.whl',
            'Pillow-7.2.0-cp37-cp37m-manylinux1_i686.whl',
            'Pillow-7.2.0-cp37-cp37m-manylinux1_x86_64.whl',
            'Pillow-7.2.0-cp37-cp37m-manylinux2014_aarch64.whl',
            'Pillow-7.2.0-cp37-cp37m-win32.whl',
            'Pillow-7.2.0-cp37-cp37m-win_amd64.whl',
            'Pillow-7.2.0-cp38-cp38-macosx_10_10_x86_64.whl',
            'Pillow-7.2.0-cp38-cp38-manylinux1_i686.whl',
            'Pillow-7.2.0-cp38-cp38-manylinux1_x86_64.whl',
            'Pillow-7.2.0-cp38-cp38-manylinux2014_aarch64.whl',
            'Pillow-7.2.0-cp38-cp38-win32.whl',
            'Pillow-7.2.0-cp38-cp38-win_amd64.whl',
            'Pillow-7.2.0-pp36-pypy36_pp73-macosx_10_10_x86_64.whl',
            'Pillow-7.2.0-pp36-pypy36_pp73-manylinux2010_x86_64.whl',
            'Pillow-7.2.0-pp36-pypy36_pp73-win32.whl',
            'reportlab-3.5.47-cp27-cp27m-macosx_10_9_x86_64.whl',
            'reportlab-3.5.47-cp27-cp27m-manylinux1_i686.whl',
            'reportlab-3.5.47-cp27-cp27m-manylinux1_x86_64.whl',
            'reportlab-3.5.47-cp27-cp27m-manylinux2010_i686.whl',
            'reportlab-3.5.47-cp27-cp27m-manylinux2010_x86_64.whl',
            'reportlab-3.5.47-cp27-cp27mu-manylinux1_i686.whl',
            'reportlab-3.5.47-cp27-cp27mu-manylinux1_x86_64.whl',
            'reportlab-3.5.47-cp27-cp27mu-manylinux2010_i686.whl',
            'reportlab-3.5.47-cp27-cp27mu-manylinux2010_x86_64.whl',
            'reportlab-3.5.47-cp27-cp27m-win32.whl',
            'reportlab-3.5.47-cp27-cp27m-win_amd64.whl',
            'reportlab-3.5.47-cp35-cp35m-macosx_10_6_intel.whl',
            'reportlab-3.5.47-cp35-cp35m-manylinux1_i686.whl',
            'reportlab-3.5.47-cp35-cp35m-manylinux1_x86_64.whl',
            'reportlab-3.5.47-cp35-cp35m-manylinux2010_i686.whl',
            'reportlab-3.5.47-cp35-cp35m-manylinux2010_x86_64.whl',
            'reportlab-3.5.47-cp35-cp35m-win32.whl',
            'reportlab-3.5.47-cp35-cp35m-win_amd64.whl',
            'reportlab-3.5.47-cp36-cp36m-macosx_10_9_x86_64.whl',
            'reportlab-3.5.47-cp36-cp36m-manylinux1_i686.whl',
            'reportlab-3.5.47-cp36-cp36m-manylinux1_x86_64.whl',
            'reportlab-3.5.47-cp36-cp36m-manylinux2010_i686.whl',
            'reportlab-3.5.47-cp36-cp36m-manylinux2010_x86_64.whl',
            'reportlab-3.5.47-cp36-cp36m-win32.whl',
            'reportlab-3.5.47-cp36-cp36m-win_amd64.whl',
            'reportlab-3.5.47-cp37-cp37m-macosx_10_9_x86_64.whl',
            'reportlab-3.5.47-cp37-cp37m-manylinux1_i686.whl',
            'reportlab-3.5.47-cp37-cp37m-manylinux1_x86_64.whl',
            'reportlab-3.5.47-cp37-cp37m-manylinux2010_i686.whl',
            'reportlab-3.5.47-cp37-cp37m-manylinux2010_x86_64.whl',
            'reportlab-3.5.47-cp37-cp37m-win32.whl',
            'reportlab-3.5.47-cp37-cp37m-win_amd64.whl',
            'reportlab-3.5.47-cp38-cp38-macosx_10_9_x86_64.whl',
            'reportlab-3.5.47-cp38-cp38-manylinux1_i686.whl',
            'reportlab-3.5.47-cp38-cp38-manylinux1_x86_64.whl',
            'reportlab-3.5.47-cp38-cp38-manylinux2010_i686.whl',
            'reportlab-3.5.47-cp38-cp38-manylinux2010_x86_64.whl',
            'reportlab-3.5.47-cp38-cp38-win32.whl',
            'reportlab-3.5.47-cp38-cp38-win_amd64.whl',
        ]
        for fn in filter(is_compatible, wheel_names):
            w = Wheel(fn)
            our_arch = ARCH.replace('linux_', '')
            for pyver, abi, arch in w.tags:
                self.assertEqual(pyver, IMPVER)
                self.assertEqual(abi, ABI)
                if sys.platform != 'darwin':
                    self.assertTrue(arch.endswith(our_arch))
                else:
                    m = re.match(r'(\w+)_(\d+)_(\d+)_(\w+)$', our_arch)
                    self.assertTrue(m)
                    _, major, minor, our_arch_kind = m.groups()
                    our_major = int(major)
                    our_minor = int(minor)
                    m = re.match(r'(\w+)_(\d+)_(\d+)_(\w+)$', arch)
                    self.assertTrue(m)
                    _, major, minor, arch_kind = m.groups()
                    major = int(major)
                    minor = int(minor)
                    self.assertEqual(major, our_major)
                    self.assertLessEqual(minor, our_minor)
                    if arch_kind in ('x86_64', 'i386'):
                        self.assertEqual(arch_kind, our_arch_kind)
                    elif arch_kind == 'fat':
                        self.assertIn(our_arch_kind, ('i386', 'ppc'))
                    elif arch_kind == 'fat3':
                        self.assertIn(our_arch_kind, ('i386', 'ppc', 'x86_x64'))
                    elif arch_kind == 'fat64':
                        self.assertIn(our_arch_kind, ('ppc64', 'x86_x64'))
                    elif arch_kind == 'intel':
                        self.assertIn(our_arch_kind, ('i386', 'x86_x64'))
                    elif arch_kind == 'universal':
                        self.assertIn(our_arch_kind, ('i386', 'ppc', 'ppc64', 'x86_x64', 'intel'))
                if 'manylinux' in arch:
                    self.assertTrue(sys.platform.startswith('linux'))
                    parts = _get_glibc_version()
                    self.assertEqual(len(parts), 2)
                    if 'manylinux2014_' in arch:
                        self.assertTrue(parts >= (2, 17))
                    if 'manylinux2010_' in arch:
                        self.assertTrue(parts >= (2, 12))
                    if 'manylinux1_' in arch:
                        self.assertTrue(parts >= (2, 5))
                    if 'manylinux_' in arch:
                        s = 'manylinux_%s_%s_' % parts
                        self.assertIn(s, arch)

    def test_is_compatible_limited_abi(self):
        major_version = sys.version_info.major
        minor_version = sys.version_info.minor
        minimum_abi3_version = (3, 2)
        if not ((major_version, minor_version) >= minimum_abi3_version and IMP_PREFIX == 'cp'):
            self.skipTest('Python %s does not support the limited API' % VER_SUFFIX)

        compatible_wheel_filenames = [
            'dummy-0.1-cp%d%d-abi3-%s.whl' % (major_version, current_minor_version, ARCH)
            for current_minor_version in range(minor_version, -1, -1)
            if (major_version, current_minor_version) >= minimum_abi3_version
        ]
        incompatible_wheel_filenames = [
            'dummy-0.1-cp%d%d-%s-%s.whl' % (major_version, current_minor_version, ABI, ARCH)
            for current_minor_version in range(minor_version - 1, -1, -1)
        ]
        for wheel_filename in compatible_wheel_filenames:
            self.assertTrue(is_compatible(wheel_filename), msg=wheel_filename)
        for wheel_filename in incompatible_wheel_filenames:
            self.assertFalse(is_compatible(wheel_filename), msg=wheel_filename)

    def test_metadata(self):
        fn = os.path.join(HERE, 'dummy-0.1-py27-none-any.whl')
        w = Wheel(fn)
        md = w.metadata
        self.assertEqual(md.name, 'dummy')
        self.assertEqual(md.version, '0.1')

    def test_invalid(self):
        fn = os.path.join(HERE, 'dummy-0.1-py27-none-any.whl')
        w = Wheel(fn)
        self.assertRaises(DistlibException, w.get_hash, b'', 'badalgo')

    def check_built_wheel(self, wheel, expected):
        for key in expected:
            self.assertEqual(expected[key], getattr(wheel, key))
        fn = os.path.join(wheel.dirname, wheel.filename)
        self.assertTrue(os.path.exists(fn))
        name, version = wheel.name, wheel.version
        with ZipFile(fn, 'r') as zf:
            for key in ('scripts', 'headers', 'data'):
                arcname = '%s-%s.data/%s/%s_file.txt' % (name, version, key, key)
                with zf.open(arcname) as bf:
                    data = bf.read()
                expected = ('dummy data - %s' % key).encode('utf-8')
                if key == 'scripts':
                    expected = b'#!python\n' + expected
                self.assertTrue(data, expected)

    def check_built_metadata(self, wheel):
        # Check the metadata of the built wheel (see #238).
        name, version = wheel.name, wheel.version
        fn = os.path.join(wheel.dirname, wheel.filename)
        self.assertTrue(os.path.exists(fn))
        arcname = '%s-%s.dist-info/WHEEL' % (name, version)
        with ZipFile(fn, 'r') as zf:
            with zf.open(arcname) as bf:
                data = bf.read().decode('utf-8')
        if wheel.arch[0] == 'any':
            is_pure = 'true'
        else:
            is_pure = 'false'
        expected = {
            'Wheel-Version': '%d.%d' % wheel.wheel_version,
            'Generator': 'distlib %s' % __version__,
            'Root-Is-Purelib': is_pure,
            'Build': '1foobar',
            'Tag': '%s-%s-%s' % (wheel.pyver[0], wheel.abi[0], wheel.arch[0]),
        }
        actual = {}
        for line in data.splitlines():
            i = line.find(':')
            key, value = line[:i], line[i + 1:].lstrip()
            actual[key] = value
        self.assertEqual(actual, expected)

    def test_build_tags(self):
        workdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, workdir)
        name = 'dummy'
        version = '0.1'
        paths = {'prefix': workdir}
        for key in ('purelib', 'platlib', 'headers', 'scripts', 'data'):
            paths[key] = p = os.path.join(workdir, key)
            os.makedirs(p)
            fn = os.path.join(p, '%s_file.txt' % key)
            with open(fn, 'w') as f:
                f.write('dummy data - %s' % key)
            if key in ('purelib', 'platlib'):
                p = os.path.join(p, '%s-%s.dist-info' % (name, version))
                os.makedirs(p)
                fn = os.path.join(p, 'RECORD')

        purelib = paths.pop('purelib')
        platlib = paths.pop('platlib')

        # Make a pure wheel with default tags
        paths['purelib'] = purelib
        wheel = Wheel('%s-%s' % (name, version))
        wheel.dirname = workdir
        wheel.build(paths)
        expected = {
            'name': name,
            'version': version,
            'pyver': [PYVER],
            'abi': ['none'],
            'arch': ['any'],
            'filename': 'dummy-0.1-%s-none-any.whl' % PYVER,
        }
        self.check_built_wheel(wheel, expected)
        # Make a pure wheel with custom tags
        pyver = [PYVER[:-1], PYVER]
        wheel.build(paths, {'pyver': pyver})
        expected = {
            'name': name,
            'version': version,
            'pyver': pyver,
            'abi': ['none'],
            'arch': ['any'],
            'filename': 'dummy-0.1-%s-none-any.whl' % '.'.join(pyver),
        }
        self.check_built_wheel(wheel, expected)

        # Make a non-pure wheel with default tags
        # On this last build in the test, set a buildver and check the metadata
        paths.pop('purelib')
        paths['platlib'] = platlib
        wheel.buildver = '1foobar'
        wheel.build(paths)
        expected['pyver'] = [IMPVER]
        expected['abi'] = [ABI]
        expected['arch'] = [ARCH]
        expected['filename'] = 'dummy-0.1-%s-%s-%s-%s.whl' % (wheel.buildver, IMPVER, ABI, ARCH)
        self.check_built_wheel(wheel, expected)
        self.check_built_metadata(wheel)

    def do_build_and_install(self, dist):
        srcdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, srcdir)
        dstdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, dstdir)

        paths = install_dist(dist, srcdir)
        paths['prefix'] = srcdir
        w = Wheel()
        w.name = paths.pop('name')
        w.version = paths.pop('version')
        w.dirname = srcdir
        pathname = w.build(paths)
        self.assertTrue(os.path.exists(pathname))

        paths = {'prefix': dstdir}
        for key in ('purelib', 'platlib', 'headers', 'scripts', 'data'):
            paths[key] = os.path.join(dstdir, key)
        w = Wheel(pathname)
        maker = ScriptMaker(None, None, add_launchers=False)
        maker.executable = os.path.join(paths['scripts'], 'python')
        dist = w.install(paths, maker)
        self.assertIsNotNone(dist)
        self.assertEqual(dist.name, w.name)
        self.assertEqual(dist.version, w.version)
        shared = dist.shared_locations
        self.assertTrue(shared)
        os.remove(pathname)
        sm = Manifest(srcdir)
        sm.findall()
        sfiles = set([os.path.relpath(p, srcdir) for p in sm.allfiles])
        dm = Manifest(dstdir)
        dm.findall()
        dfiles = set([os.path.relpath(p, dstdir) for p in dm.allfiles])
        omitted = sfiles - dfiles
        omitted = omitted.pop()
        endings = os.path.join('.dist-info', 'WHEEL'), '.pyc', '.pyo'
        self.assertTrue(omitted.endswith(endings))

    def test_version_incompatibility(self):

        class Warner(object):

            def __call__(self, wheel_version, file_version):
                self.wheel_version = wheel_version
                self.file_version = file_version

        fn = os.path.join(HERE, 'dummy-0.1-py27-none-any.whl')
        dstdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, dstdir)
        w = Wheel(fn)
        paths = {'prefix': dstdir}
        for key in ('purelib', 'platlib', 'headers', 'scripts', 'data'):
            paths[key] = os.path.join(dstdir, key)
        warner = Warner()
        maker = ScriptMaker(None, None)
        w.install(paths, maker, warner=warner)
        self.assertEqual(warner.wheel_version, w.wheel_version)
        self.assertEqual(warner.file_version, (2, 0))
        # Now set the wheel's instance to the higher value and ensure
        # warner isn't called
        warner = Warner()
        w.wheel_version = (2, 0)
        w.install(paths, maker, warner=warner)
        self.assertFalse(hasattr(warner, 'wheel_version'))
        self.assertFalse(hasattr(warner, 'file_version'))

    def test_custom_executable(self):
        fn = os.path.join(HERE, 'dummy-0.1-py27-none-any.whl')
        for executable in 'mypython', None:
            dstdir = tempfile.mkdtemp()
            self.addCleanup(shutil.rmtree, dstdir)
            w = Wheel(fn)
            paths = {'prefix': dstdir}
            for key in ('purelib', 'platlib', 'headers', 'scripts', 'data'):
                paths[key] = os.path.join(dstdir, key)
            maker = ScriptMaker(None, None)
            maker.variants = set([''])
            maker.executable = executable
            w.install(paths, maker)
            # On Windows there will be an exe file, and on POSIX a text file.
            # The test is structured to not care.
            p = paths['scripts']
            # there should be just one file in the directory - dummy.py/dummy.exe
            p = os.path.join(p, os.listdir(p)[0])
            with open(p, 'rb') as f:
                data = f.read()
            if executable is None:
                expected = fsencode(get_executable())
            else:
                expected = executable.encode('utf-8')
            expected = b'#!' + expected + b' -E'
            if not sysconfig.is_python_build():
                self.assertIn(expected, data)

    def test_verify(self):
        fn = os.path.join(HERE, 'dummy-0.1-py27-none-any.whl')
        w = Wheel(fn)
        w.verify()
        # see issue 115
        fn = os.path.join(HERE, 'valid_wheel-0.0.1-py3-none-any.whl')
        w = Wheel(fn)
        w.verify()
        fn = os.path.join(HERE, 'bad_wheels', 'dummy-0.1-py27-none-any.whl')
        w = Wheel(fn)
        self.assertRaises(DistlibException, w.verify)

    def wheel_modifier_nop(self, path_map):
        return False

    def wheel_modifier(self, path_map):
        mdpath = path_map['dummy-0.1.dist-info/%s' % LEGACY_METADATA_FILENAME]
        md = Metadata(path=mdpath)
        md.add_requirements(['numpy'])
        md.write(path=mdpath, legacy=True)
        return True

    def wheel_modifier_ver(self, path_map):
        mdpath = path_map['dummy-0.1.dist-info/%s' % LEGACY_METADATA_FILENAME]
        md = Metadata(path=mdpath)
        md.version = '0.1+123'
        md.write(path=mdpath, legacy=True)
        return True

    def test_update(self):
        workdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, workdir)
        fn = 'dummy-0.1-py27-none-any.whl'
        sfn = os.path.join(HERE, fn)
        dfn = os.path.join(workdir, fn)
        shutil.copyfile(sfn, dfn)
        mtime = os.stat(dfn).st_mtime
        w = Wheel(dfn)
        modified = w.update(self.wheel_modifier_nop)
        self.assertFalse(modified)
        self.assertEqual(mtime, os.stat(dfn).st_mtime)
        modified = w.update(self.wheel_modifier)
        self.assertTrue(modified)
        self.assertLessEqual(mtime, os.stat(dfn).st_mtime)
        w = Wheel(dfn)
        w.verify()
        md = w.metadata
        self.assertEqual(md.run_requires, ['numpy'])
        self.assertEqual(md.version, '0.1+1')

        modified = w.update(self.wheel_modifier_ver)
        self.assertTrue(modified)
        self.assertLessEqual(mtime, os.stat(dfn).st_mtime)
        w = Wheel(dfn)
        w.verify()
        md = w.metadata
        self.assertEqual(md.run_requires, ['numpy'])
        self.assertEqual(md.version, '0.1+123')

    def test_info(self):
        fn = os.path.join(HERE, 'dummy-0.1-py27-none-any.whl')
        w = Wheel(fn)
        actual = w.info
        actual.pop('Generator', None)
        expected = {'Root-Is-Purelib': 'true', 'Tag': 'py27-none-any', 'Wheel-Version': '2.0'}
        self.assertEqual(actual, expected)

    @unittest.skipIf(sys.version_info[:2] != (2, 7), 'The test wheel is only '
                     '2.7 mountable')
    def test_mount(self):
        fn = os.path.join(HERE, 'dummy-0.1-py27-none-any.whl')
        w = Wheel(fn)
        self.assertNotIn(fn, sys.path)
        w.mount()
        self.assertIn(fn, sys.path)
        w.unmount()
        self.assertNotIn(fn, sys.path)

    def test_mount_extensions(self):
        if PYVER == 'py27':
            fn = 'minimext-0.1-cp27-none-linux_x86_64.whl'
        elif PYVER == 'py32':
            fn = 'minimext-0.1-cp32-cp32mu-linux_x86_64.whl'
        elif PYVER == 'py33':
            fn = 'minimext-0.1-cp33-cp33m-linux_x86_64.whl'
        else:
            fn = None
        if not fn:  # pragma: no cover
            raise unittest.SkipTest('Suitable wheel not found.')
        fn = os.path.join(HERE, fn)
        w = Wheel(fn)
        if not w.is_compatible() or not w.is_mountable():  # pragma: no cover
            raise unittest.SkipTest('Wheel not suitable for mounting.')
        self.assertRaises(ImportError, __import__, 'minimext')
        w.mount()
        mod = __import__('minimext')
        self.assertIs(mod, sys.modules['minimext'])
        self.assertEqual(mod.fib(10), 55)
        w.unmount()
        del sys.modules['minimext']
        self.assertRaises(ImportError, __import__, 'minimext')

    def test_local_version(self):
        w = Wheel('dummy-0.1_1.2')
        self.assertEqual(w.filename, 'dummy-0.1_1.2-%s'
                         '-none-any.whl' % PYVER)
        self.assertEqual(w.name, 'dummy')
        self.assertEqual(w.version, '0.1-1.2')
        self.assertFalse(w.exists)
        w.version = '0.1-1.3'
        self.assertEqual(w.filename, 'dummy-0.1_1.3-%s'
                         '-none-any.whl' % PYVER)

    def test_abi(self):
        pyver = sysconfig.get_config_var('py_version_nodot')
        if not pyver:
            pyver = '%s%s' % sys.version_info[:2]
        parts = ['cp', pyver]
        if sysconfig.get_config_var('Py_DEBUG'):
            parts.append('d')
        # Starting with 3.8, the SOABI doesn't append m when WITH_PYMALLOC is
        # defined (see bpo-36707)
        vi = sys.version_info[:2]
        if IMP_PREFIX == 'cp':
            if vi < (3, 8):
                wpm = sysconfig.get_config_var('WITH_PYMALLOC')
                if wpm or wpm is None:
                    parts.append('m')
            if vi < (3, 3):
                us = sysconfig.get_config_var('Py_UNICODE_SIZE')
                if us == 4 or (us is None and sys.maxunicode == 0x10FFFF):
                    parts.append('u')
            if bool(sysconfig.get_config_var("Py_GIL_DISABLED")):
                parts.append('t')
        if vi < (3, 5):
            abi = ABI
        else:
            abi = ABI.split('-')[0]
        self.assertEqual(''.join(parts), abi)

    @unittest.skipIf('SKIP_ONLINE' in os.environ, 'Skipping online test')
    @unittest.skipUnless(PIP_AVAILABLE, 'pip is needed for this test')
    @unittest.skipIf(sys.version_info[:2] >= (3, 7), 'The test distribution is not '
                     '3.7+ compatible')
    def test_build_and_install_pure(self):
        if pip_version() >= (20, 2, 0):
            raise unittest.SkipTest('Test not supported by pip version')
        self.do_build_and_install('sarge == 0.1')

    @unittest.skipIf('SKIP_ONLINE' in os.environ, 'Skipping online test')
    @unittest.skipIf(hasattr(sys, 'pypy_version_info'), 'The test distribution'
                     ' does not build on PyPy')
    @unittest.skipIf(sys.platform != 'linux2', 'The test distribution only '
                     'builds on Linux')
    @unittest.skipUnless(PIP_AVAILABLE, 'pip is needed for this test')
    def test_build_and_install_plat(self):
        if pip_version() >= (20, 2, 0):
            raise unittest.SkipTest('Test not supported by pip version')
        self.do_build_and_install('hiredis == 0.1.1')

    @unittest.skipIf('SKIP_ONLINE' in os.environ, 'Skipping online test')
    @unittest.skipIf(sys.version_info[0] == 3, 'The test distribution is not '
                     '3.x compatible')
    @unittest.skipUnless(PIP_AVAILABLE, 'pip is needed for this test')
    def test_build_and_install_data(self):
        if pip_version() >= (20, 2, 0):
            raise unittest.SkipTest('Test not supported by pip version')
        self.do_build_and_install('Werkzeug == 0.5')

    @unittest.skipIf('SKIP_ONLINE' in os.environ, 'Skipping online test')
    @unittest.skipIf(sys.version_info[0] == 3, 'The test distribution is not '
                     '3.x compatible')
    @unittest.skipUnless(PIP_AVAILABLE, 'pip is needed for this test')
    def test_build_and_install_scripts(self):
        if pip_version() >= (20, 2, 0):
            raise unittest.SkipTest('Test not supported by pip version')
        self.do_build_and_install('Babel == 0.9.6')


if __name__ == '__main__':  # pragma: no cover
    import logging
    logging.basicConfig(level=logging.DEBUG, filename='test_wheel.log', filemode='w', format='%(message)s')
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2013 Vinay Sajip.
# Licensed to the Python Software Foundation under a contributor agreement.
# See LICENSE.txt and CONTRIBUTORS.txt.
#
from __future__ import unicode_literals

from operator import attrgetter
import os
import sys

from compat import unittest
from support import DistlibTestCase, in_github_workflow

from distlib import DistlibException
from distlib.resources import finder, finder_for_path, ResourceCache
from distlib.util import get_cache_base

HERE = os.path.abspath(os.path.dirname(__file__))
IN_GITHUB_WORKFLOW = in_github_workflow()


class ZipResourceTestCase(DistlibTestCase):

    def setUp(self):
        sys.path.insert(0, os.path.join(HERE, 'foo.zip'))

    def tearDown(self):
        sys.path.pop(0)

    def test_existing_resource(self):
        f = finder('foo')
        r = f.find('foo_resource.bin')
        self.assertTrue(r)
        self.assertEqual(r.bytes, b'more_data\n')
        self.assertEqual(r.size, 10)
        stream = r.as_stream()
        self.assertEqual(stream.read(), b'more_data\n')
        stream.close()
        # can access subpackage bar's resources using subdir path ...
        r = f.find('bar/bar_resource.bin')
        self.assertTrue(r)
        self.assertFalse(r.is_container)
        self.assertEqual(r.bytes, b'data\n')
        self.assertEqual(r.size, 5)
        stream = r.as_stream()
        self.assertEqual(stream.read(), b'data\n')
        stream.close()

        r = f.find('bar')
        self.assertTrue(r)
        self.assertTrue(r.is_container)
        self.assertRaises(AttributeError, attrgetter('bytes'), r)
        self.assertRaises(AttributeError, attrgetter('size'), r)
        self.assertRaises(AttributeError, attrgetter('file_path'), r)
        f = finder('foo.bar')
        r = f.find('bar_resource.bin')
        self.assertTrue(r)
        self.assertEqual(r.bytes, b'data\n')
        self.assertEqual(r.size, 5)
        stream = r.as_stream()
        self.assertEqual(stream.read(), b'data\n')
        stream.close()

    def test_nonexistent_resource(self):
        f = finder('foo')
        r = f.find('no_such_resource.bin')
        self.assertIsNone(r)

    def test_non_package(self):
        self.assertRaises(DistlibException, finder, 'foo.bar.baz')

    def test_contents(self):
        f = finder('foo')
        r = f.find('foo_resource.bin')
        self.assertTrue(r)
        self.assertRaises(AttributeError, attrgetter('resources'), r)
        r = f.find('bar')
        self.assertTrue(r)
        expected = set(('bar_resource.bin', 'baz.py', '__init__.py'))
        self.assertEqual(r.resources, expected)

    def test_root_resources(self):
        f = finder('foo')
        r = f.find('')
        self.assertTrue(r)
        self.assertTrue(r.is_container)
        expected = set(('foo_resource.bin', 'bar', '__init__.py'))
        self.assertEqual(r.resources, expected)

    def test_dir_in_zip(self):
        sys.path[0] = '%s/lib' % os.path.join(HERE, 'bar.zip')
        f = finder('barbar')
        self.assertIsNone(f.find('readme.txt'))
        r = f.find('bar_resource.bin')
        self.assertTrue(r)
        self.assertFalse(r.is_container)
        f = finder('barbar.baz')
        r = f.find('baz_resource.bin')
        self.assertTrue(r)
        self.assertFalse(r.is_container)

    def test_finder_for_path(self):
        f = finder_for_path(sys.path[0])
        r = f.find('')
        self.assertIsNotNone(r)
        self.assertTrue(r.is_container)
        p = os.path.join(sys.path[0], 'foo')
        f = finder_for_path(p)
        r = f.find('')
        self.assertIsNotNone(r)
        self.assertTrue(r.is_container)

    def test_iterator(self):
        f = finder('foo')
        iterator = f.iterator('')
        actual = set([(r.name, r.is_container) for r in iterator])
        expected = set([('', True), ('foo_resource.bin', False), ('bar/bar_resource.bin', False), ('bar/baz.py', False),
                        ('__init__.py', False), ('bar', True), ('bar/__init__.py', False)])
        self.assertEqual(actual, expected)
        iterator = f.iterator('bar')
        actual = set([(r.name, r.is_container) for r in iterator])
        expected = set([('bar/baz.py', False), ('bar', True), ('bar/bar_resource.bin', False),
                        ('bar/__init__.py', False)])
        self.assertEqual(actual, expected)
        iterator = f.iterator('bar/bar_resource.bin')
        actual = set([(r.name, r.is_container) for r in iterator])
        self.assertEqual(actual, set([('bar/bar_resource.bin', False)]))


class FileResourceTestCase(DistlibTestCase):

    def setUp(self):
        sys.path.insert(0, HERE)

    def tearDown(self):
        sys.path.pop(0)

    @unittest.skipIf(IN_GITHUB_WORKFLOW, 'This test is end-of-line dependent')
    def test_existing_resource(self):
        f = finder('foofoo')
        r = f.find('foo_resource.bin')
        self.assertTrue(r)
        self.assertFalse(r.is_container)
        self.assertEqual(r.bytes, b'more_data\n')
        self.assertEqual(r.size, 10)
        stream = r.as_stream()
        self.assertEqual(stream.read(), b'more_data\n')
        stream.close()
        r = f.find('bar')
        self.assertTrue(r)
        self.assertTrue(r.is_container)
        self.assertRaises(AttributeError, attrgetter('bytes'), r)
        self.assertRaises(AttributeError, attrgetter('size'), r)
        f = finder('foofoo.bar')
        r = f.find('bar_resource.bin')
        self.assertTrue(r)
        self.assertFalse(r.is_container)
        self.assertEqual(r.bytes, b'data\n')
        self.assertEqual(r.size, 5)
        stream = r.as_stream()
        self.assertEqual(stream.read(), b'data\n')
        stream.close()

    def test_nonexistent_resource(self):
        f = finder('foofoo')
        r = f.find('no_such_resource.bin')
        self.assertIsNone(r)

    def test_contents(self):
        f = finder('foofoo')
        r = f.find('foo_resource.bin')
        self.assertTrue(r)
        self.assertRaises(AttributeError, attrgetter('resources'), r)
        r = f.find('bar')
        self.assertTrue(r)
        expected = set(('bar_resource.bin', 'baz.py', '__init__.py'))
        self.assertEqual(r.resources, expected)

    def test_root_resources(self):
        f = finder('foofoo')
        r = f.find('')
        self.assertTrue(r)
        self.assertTrue(r.is_container)
        expected = set(('foo_resource.bin', 'bar', '__init__.py', 'nested'))
        self.assertEqual(r.resources, expected)

    @unittest.skipIf(IN_GITHUB_WORKFLOW, 'This test is end-of-line dependent')
    def test_nested(self):
        f = finder('foofoo')
        r = f.find('nested/nested_resource.bin')
        self.assertTrue(r)
        self.assertFalse(r.is_container)
        self.assertEqual(r.bytes, b'nested data\n')
        stream = r.as_stream()
        self.assertEqual(stream.read(), b'nested data\n')
        stream.close()
        r = f.find('nested')
        self.assertTrue(r)
        self.assertTrue(r.is_container)
        self.assertTrue(r)
        self.assertEqual(r.resources, set(['nested_resource.bin']))

    @unittest.skipIf(sys.version_info[0] != 2, 'This test on Python 2 only')
    def test_bytes_path(self):
        f = finder('foofoo')
        for path in 'foo/b\xe7r', b'foo/b\xe7r':
            self.assertEqual(type(f._make_path(path)), type(path))

    def test_iterator(self):
        f = finder('foofoo')
        iterator = f.iterator('')
        actual = set([(r.name, r.is_container) for r in iterator])
        expected = set([
            ('', True),
            ('nested/nested_resource.bin', False),
            ('bar', True),
            ('__init__.py', False),
            ('nested', True),
            ('bar/bar_resource.bin', False),
            ('bar/__init__.py', False),
            ('bar/baz.py', False),
            ('foo_resource.bin', False),
        ])
        self.assertEqual(actual, expected)
        iterator = f.iterator('bar')
        actual = set([(r.name, r.is_container) for r in iterator])
        expected = set([('bar/baz.py', False), ('bar', True), ('bar/bar_resource.bin', False),
                        ('bar/__init__.py', False)])
        self.assertEqual(actual, expected)
        iterator = f.iterator('bar/bar_resource.bin')
        actual = set([(r.name, r.is_container) for r in iterator])
        self.assertEqual(actual, set([('bar/bar_resource.bin', False)]))


class CacheTestCase(DistlibTestCase):

    def test_base(self):
        cache = ResourceCache()
        expected = os.path.join(get_cache_base(), str('resource-cache'))
        self.assertEqual(expected, cache.base)
        self.assertTrue(os.path.isdir(expected))

    def test_filepath(self):
        path = os.path.join(HERE, 'foo.zip')
        sys.path.insert(0, path)
        self.addCleanup(sys.path.remove, path)
        sys.path.insert(0, HERE)
        self.addCleanup(sys.path.remove, HERE)
        path = '%s/lib' % os.path.join(HERE, 'bar.zip')
        sys.path.insert(0, path)
        self.addCleanup(sys.path.remove, path)

        cases = (('foo', 'foo_resource.bin'), ('foo', 'bar/bar_resource.bin'), ('foofoo', 'bar/bar_resource.bin'),
                 ('barbar', 'bar_resource.bin'), ('barbar.baz', 'baz_resource.bin'))

        for pkg, path in cases:
            f = finder(pkg)
            r = f.find(path)
            fp = r.file_path
            with open(fp, 'rb') as df:
                data = df.read()
            self.assertEqual(data, r.bytes)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 Vinay Sajip.
# Licensed to the Python Software Foundation under a contributor agreement.
# See LICENSE.txt and CONTRIBUTORS.txt.
#
from __future__ import unicode_literals

import os
import shutil
import subprocess
import sys
import tempfile
import textwrap
try:
    import venv
except ImportError:
    venv = None

from compat import unittest
from support import DistlibTestCase

from distlib.compat import fsencode, sysconfig
from distlib.scripts import ScriptMaker, enquote_executable
from distlib.util import get_executable

HERE = os.path.abspath(os.path.dirname(__file__))

COPIED_SCRIPT = '''#!python
# This is a copied script
'''

MADE_SCRIPT = 'made = dummy.module:main'


class ScriptTestCase(DistlibTestCase):

    def setUp(self):
        source_dir = os.path.join(HERE, 'scripts')
        target_dir = tempfile.mkdtemp(prefix='distlib-test-')
        self.maker = ScriptMaker(source_dir, target_dir, add_launchers=False)

    def tearDown(self):
        shutil.rmtree(self.maker.target_dir)

    @unittest.skipIf(sysconfig.is_python_build(), 'Test not appropriate for '
                     'Python source builds')
    def test_shebangs(self):
        executable = fsencode(get_executable())
        for fn in ('foo.py', 'script1.py', 'script2.py', 'script3.py', 'shell.sh'):
            files = self.maker.make(fn)
            self.assertEqual(len(files), 1)
            d, f = os.path.split(files[0])
            self.assertEqual(f, fn)
            self.assertEqual(d, self.maker.target_dir)
            if fn.endswith('.py') and fn != 'foo.py':  # no shebang in foo.py
                with open(files[0], 'rb') as f:
                    first_line = f.readline()
                self.assertIn(executable, first_line)

    def test_shebangs_custom_executable(self):
        srcdir = tempfile.mkdtemp(prefix='distlib-test-')
        self.addCleanup(shutil.rmtree, srcdir)
        dstdir = tempfile.mkdtemp(prefix='distlib-test-')
        self.addCleanup(shutil.rmtree, dstdir)
        maker = ScriptMaker(srcdir, dstdir, add_launchers=False)
        maker.executable = 'this_should_appear_in_the_shebang_line(中文)'
        # let's create the script to be copied. It has a vanilla shebang line,
        # with some Unicode in it.
        fn = os.path.join(srcdir, 'copied')
        with open(fn, 'w') as f:
            f.write(COPIED_SCRIPT)
        # Let's ask the maker to copy the script, and see what the shebang is
        # in the copy.
        filenames = maker.make('copied')
        with open(filenames[0], 'rb') as f:
            actual = f.readline().decode('utf-8')
        self.assertIn(maker.executable, actual)
        # Now let's make a script from a callable
        filenames = maker.make(MADE_SCRIPT)
        with open(filenames[0], 'rb') as f:
            actual = f.readline().decode('utf-8')
        self.assertIn(maker.executable, actual)

    @unittest.skipIf(os.name != 'posix', 'Test only appropriate for '
                     'POSIX systems')
    def test_custom_shebang(self):
        # Construct an executable with a space in it
        self.maker.executable = 'an executable with spaces'
        filenames = self.maker.make('script1.py')
        with open(filenames[0], 'rb') as f:
            first_line = f.readline()
            second_line = f.readline()
            third_line = f.readline()
        self.assertEqual(first_line, b'#!/bin/sh\n')
        self.assertEqual(second_line, b"'''exec' an executable with "
                         b'spaces "$0" "$@"\n')
        self.assertEqual(third_line, b"' '''\n")
        # Python 3.3 cannot create a venv in an existing directory
        if venv and sys.version_info[:2] >= (3, 4):
            if sys.platform == 'darwin':
                # Supposedly 512, but various symlinks mean that temp folder
                # names get larger than you'd expect ... might vary on different
                # OS versions, too
                dlen = 220
            else:
                dlen = 127
            dstdir = tempfile.mkdtemp(suffix='cataaaaaa' + 'a' * dlen)
            self.addCleanup(shutil.rmtree, dstdir)
            bindir = os.path.join(dstdir, 'bin')
            maker = ScriptMaker(self.maker.source_dir, bindir, add_launchers=False)
            # See issue #204. Changed to use symlinks on POSIX
            venv.create(dstdir, symlinks=(os.name != 'nt'))
            maker.executable = os.path.join(bindir, 'python')
            filenames = maker.make('script8.py')
            p = subprocess.Popen(filenames[0], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()
            self.assertEqual(p.returncode, 0)
            self.assertEqual(stderr, b'')
            expected = os.path.realpath(maker.executable)  # symlinks on OS X
            actual = os.path.realpath(stdout.strip())
            self.assertEqual(actual, expected.encode('utf-8'))

    @unittest.skipIf(os.name != 'posix', 'Test only appropriate for '
                     'POSIX systems')
    def test_cross_env_shebang(self):
        # Construct an executable with a path that wouldn't ordinarily be a problem
        self.maker.executable = '/path/to/python3'

        # Set the cross-compiling marker attribute, then clean up.
        try:
            sys.cross_compiling = True
            filenames = self.maker.make('script1.py')
        finally:
            delattr(sys, "cross_compiling")

        with open(filenames[0], 'rb') as f:
            first_line = f.readline()
            second_line = f.readline()
            third_line = f.readline()
        self.assertEqual(first_line, b'#!/bin/sh\n')
        self.assertEqual(second_line, b"'''exec' /path/to/python3 "
                         b'"$0" "$@"\n')
        self.assertEqual(third_line, b"' '''\n")

    def test_multiple(self):
        specs = ('foo.py', 'script1.py', 'script2.py', 'script3.py', 'shell.sh', 'uwsgi_part')
        files = self.maker.make_multiple(specs)
        self.assertEqual(len(specs), len(files))
        expected = set(specs)
        self.assertEqual(expected, set([os.path.basename(f) for f in files]))
        ofiles = os.listdir(self.maker.target_dir)
        self.assertEqual(expected, set(ofiles))

    def test_generation(self):
        self.maker.clobber = True

        def do_test(maker, version_info):
            for name in ('main', 'other_main'):
                for options in (None, {}, {'gui': False}, {'gui': True}):
                    gui = options and options.get('gui', False)
                    spec = 'foo = foo:' + name
                    files = maker.make(spec, options)
                    self.assertEqual(len(files), 2)
                    actual = set()
                    for f in files:
                        d, f = os.path.split(f)
                        actual.add(f)
                    if os.name == 'nt':  # pragma: no cover
                        if gui:
                            ext = 'pyw'
                        else:
                            ext = 'py'
                        expected = set(['foo.%s' % ext, 'foo-%s.%s.%s' % (version_info[0], version_info[1], ext)])
                    else:
                        expected = set(['foo', 'foo-%s.%s' % (version_info[0], version_info[1])])
                    self.assertEqual(actual, expected)
                    self.assertEqual(d, maker.target_dir)
                    for fn in files:
                        with open(fn, 'r') as f:
                            text = f.read()
                        # self.assertIn("_resolve('foo', '%s')" % name, text)
                        if options and options['gui'] and os.name == 'nt':  # pragma: no cover
                            first_line, rest = text.split('\n', 1)
                            self.assertIn('pythonw', first_line)

        do_test(self.maker, sys.version_info)
        # See issue 134. Test for specifying the target Python version
        self.maker.version_info = version_info = (4, 4)  # just any other Python version
        do_test(self.maker, version_info)

    def test_clobber(self):
        files = self.maker.make('foo = foo:main')
        saved_files = files
        self.assertGreaterEqual(len(files), 2)  # foo, foo-X.Y
        files = self.maker.make('foo = foo:main')
        self.assertFalse(files)
        self.maker.clobber = True
        files = self.maker.make('foo = foo:main')
        self.assertEqual(files, saved_files)

    @unittest.skipIf(os.name != 'nt', 'Test is Windows-specific')
    def test_launchers(self):  # pragma: no cover
        tlauncher = self.maker._get_launcher('t')
        self.maker.add_launchers = True
        specs = ('foo.py', 'script1.py', 'script2.py', 'script3.py', 'shell.sh')
        files = self.maker.make_multiple(specs)
        self.assertEqual(len(specs), len(files))
        filenames = set([os.path.basename(f) for f in files])
        self.assertEqual(filenames, set(('foo.py', 'script1.exe', 'script2.exe', 'script3.exe', 'shell.sh')))
        for fn in files:
            if not fn.endswith('.exe'):
                continue
            with open(fn, 'rb') as f:
                data = f.read()
            self.assertTrue(data.startswith(tlauncher))

    @unittest.skipIf(os.name != 'nt', 'Test is Windows-specific')
    def test_launcher_run(self):
        self.maker.add_launchers = True
        files = self.maker.make('script6.py')
        self.assertEqual(len(files), 1)
        p = subprocess.Popen([files[0], 'Test Argument'],
                             stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
        stdout, stderr = p.communicate('input'.encode('ascii'))
        actual = stdout.decode('ascii').replace('\r\n', '\n')
        expected = textwrap.dedent("""
            script6.exe
            ['Test Argument']
            'input'
            non-optimized
            """).lstrip()
        self.assertEqual(actual, expected)

    @unittest.skipIf(os.name != 'nt', 'Test is Windows-specific')
    def test_launcher_run_with_interpreter_args(self):
        srcdir = tempfile.mkdtemp(prefix='distlib-test-')
        self.addCleanup(shutil.rmtree, srcdir)
        dstdir = tempfile.mkdtemp(prefix='distlib-test-')
        self.addCleanup(shutil.rmtree, dstdir)
        maker = ScriptMaker(srcdir, dstdir, add_launchers=True)

        # add '-O' option to shebang to run in optimized mode
        with open(os.path.join(HERE, 'scripts', 'script6.py'), 'r') as src:
            with open(os.path.join(srcdir, 'script6-optimized.py'), 'w') as dst:
                shebang = src.readline().rstrip()
                dst.write(shebang + " -O\n")
                dst.write(src.read())

        files = maker.make('script6-optimized.py')
        self.assertEqual(len(files), 1)
        p = subprocess.Popen([files[0], 'Test Argument'],
                             stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
        stdout, stderr = p.communicate('input'.encode('ascii'))
        actual = stdout.decode('ascii').replace('\r\n', '\n')
        expected = textwrap.dedent("""
            script6-optimized.exe
            ['Test Argument']
            'input'
            """).lstrip()  # 'non-optimized' is not printed this time
        self.assertEqual(actual, expected)

    @unittest.skipIf(os.name != 'nt', 'Test is Windows-specific')
    def test_windows(self):  # pragma: no cover
        wlauncher = self.maker._get_launcher('w')
        tlauncher = self.maker._get_launcher('t')
        self.maker.add_launchers = True
        # executable = os.path.normcase(sys.executable).encode('utf-8')
        executable = sys.executable.encode('utf-8')
        if b'python3.' in executable:
            wexecutable = executable.replace(b'python3.', b'pythonw3.')
        else:
            wexecutable = executable.replace(b'python.', b'pythonw.')
        files = self.maker.make('script4.py')
        self.assertEqual(len(files), 1)
        filenames = set([os.path.basename(f) for f in files])
        self.assertEqual(filenames, set(['script4.exe']))
        for fn in files:
            with open(fn, 'rb') as f:
                data = f.read()
            self.assertTrue(data.startswith(wlauncher))
            self.assertIn(executable, data)
        # Now test making scripts gui and console
        files = self.maker.make('foo = foo:main', {'gui': True})
        self.assertEqual(len(files), 2)
        filenames = set([os.path.basename(f) for f in files])
        specific = '%s.%s' % sys.version_info[:2]
        self.assertEqual(filenames, set(('foo.exe', 'foo-%s.exe' % specific)))
        for fn in files:
            with open(fn, 'rb') as f:
                data = f.read()
            self.assertTrue(data.startswith(wlauncher))
            self.assertIn(wexecutable, data)

        files = self.maker.make('foo = foo:main')
        self.assertEqual(len(files), 2)
        filenames = set([os.path.basename(f) for f in files])
        self.assertEqual(filenames, set(('foo.exe', 'foo-%s.exe' % specific)))
        for fn in files:
            with open(fn, 'rb') as f:
                data = f.read()
            self.assertTrue(data.startswith(tlauncher))
            self.assertIn(executable, data)

    @unittest.skipIf(os.name != 'nt', 'Test is Windows-specific')
    def test_windows_run(self):
        self.maker.add_launchers = True
        files = self.maker.make('script7.pyw')
        self.assertEqual(len(files), 1)

        test_output = os.path.join(self.maker.target_dir, 'test_output.txt')
        p = subprocess.Popen([files[0], test_output, 'Test Argument'],
                             stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
        stdout, stderr = p.communicate()
        self.assertFalse(stdout)
        self.assertFalse(stderr)
        with open(test_output, 'rb') as f:
            actual = f.read().decode('ascii')
        self.assertEqual(actual, 'Test Argument')

    def test_dry_run(self):
        self.maker.dry_run = True
        self.maker.variants = set([''])
        specs = ('foo.py', 'bar = foo:main')
        files = self.maker.make_multiple(specs)
        self.assertEqual(len(specs), len(files))
        if os.name == 'nt':  # pragma: no cover
            bar = 'bar.py'
        else:
            bar = 'bar'
        self.assertEqual(set(('foo.py', bar)), set([os.path.basename(f) for f in files]))
        ofiles = os.listdir(self.maker.target_dir)
        self.assertFalse(ofiles)

    def test_script_run(self):
        if sys.version_info[:2] < (3, 13):
            target = 'cgi:print_directory'
        elif os.name != 'nt':
            target = 'test.support.interpreters:list_all'
        else:
            raise unittest.SkipTest('test not available on Windows for Python >= 3.13')
        files = self.maker.make('test = %s' % target)
        self.assertEqual(len(files), 2)
        p = subprocess.Popen([sys.executable, files[0]], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
        if sys.version_info[:2] < (3, 13):
            self.assertIn(b'<H3>Current Working Directory:</H3>', stdout)
            self.assertIn(os.getcwd().encode('utf-8'), stdout)
        else:
            self.assertIn(b'[Interpreter(0)]', stderr)
            self.assertEqual(p.returncode, 1)

    @unittest.skipUnless(os.name == 'posix', 'Test only valid for POSIX')
    def test_mode(self):
        # save test runner's original umask and ensure default 022
        saved_umask = os.umask(0o022)
        try:
            self.maker.set_mode = False
            files = self.maker.make('foo = foo:main')
            self.assertEqual(len(files), 2)
            for f in files:
                self.assertIn(os.stat(f).st_mode & 0o7777, (0o644, 0o664))
            self.maker.set_mode = True
            files = self.maker.make('bar = bar:main')
            self.assertEqual(len(files), 2)
            for f in files:
                self.assertIn(os.stat(f).st_mode & 0o7777, (0o755, 0o775))
        finally:
            # restore the test runner's original umask
            os.umask(saved_umask)

    def test_interpreter_args(self):
        executable = fsencode(get_executable())
        options = {'interpreter_args': ['-E', '"foo bar"', 'baz frobozz']}
        self.maker.variants = set([''])
        files = self.maker.make('foo = bar:baz', options=options)
        self.assertEqual(len(files), 1)
        with open(files[0], 'rb') as f:
            shebang_line = f.readline()
        if not sysconfig.is_python_build():
            self.assertIn(executable, shebang_line)
        self.assertIn(b' -E "foo bar" baz frobozz', shebang_line)

    def test_args_on_copy(self):
        self.maker.variants = set([''])
        self.maker.executable = 'mypython'
        files = self.maker.make('script5.py')
        with open(files[0]) as f:
            actual = f.readline().strip()
        self.assertEqual(actual, '#!mypython -mzippy.activate')
        if not sysconfig.is_python_build():
            self.maker.executable = None
            os.remove(files[0])
            files = self.maker.make('script5.py')
            with open(files[0]) as f:
                actual = f.readline().strip()
            expected = '#!%s -mzippy.activate' % get_executable()
            self.assertEqual(actual, expected)

    def test_enquote_executable(self):
        for executable, expected in (('/no/spaces', '/no/spaces'), ('/i have/space', '"/i have/space"'),
                                     ('"/space prequoted"', '"/space prequoted"'),
                                     ('/usr/bin/env nospaces', '/usr/bin/env nospaces'), ('/usr/bin/env with spaces',
                                                                                          '/usr/bin/env "with spaces"'),
                                     ('/usr/bin/env "pre spaced"', '/usr/bin/env "pre spaced"')):
            self.assertEqual(enquote_executable(executable), expected)

    def test_variant_configuration(self):
        spec = 'foo = foo:main'
        files = self.maker.make(spec)
        if os.name == 'nt':
            ext = '.py'
        else:
            ext = ''
        basenames = ('foo%s' % ext, 'foo-%s.%s%s' % (self.maker.version_info[:2] + (ext, )))
        expected = set([os.path.join(self.maker.target_dir, s) for s in basenames])
        self.assertEqual(expected, set(files))
        self.maker.variant_separator = ''
        self.maker.clobber = True
        files = self.maker.make(spec)
        basenames = ('foo%s' % ext, 'foo%s.%s%s' % (self.maker.version_info[:2] + (ext, )))
        expected = set([os.path.join(self.maker.target_dir, s) for s in basenames])
        self.assertEqual(expected, set(files))

        class CustomMaker(ScriptMaker):

            def get_script_filenames(self, name):
                result = super(CustomMaker, self).get_script_filenames(name)
                if 'X.Y' in self.variants:
                    result.add('%s%s.%s' % ((name, ) + self.version_info[:2]))
                return result

        maker = CustomMaker(self.maker.source_dir, self.maker.target_dir, add_launchers=False)
        maker.clobber = True
        files = maker.make(spec)
        basenames = ('foo%s' % ext, 'foo%s.%s%s' % (self.maker.version_info[:2] + (ext, )),
                     'foo-%s.%s%s' % (self.maker.version_info[:2] + (ext, )))
        expected = set([os.path.join(self.maker.target_dir, s) for s in basenames])
        self.assertEqual(expected, set(files))
        if os.name == 'nt':
            maker.add_launchers = True
            files = maker.make(spec)
            expected = set([e.replace('.py', '.exe') for e in expected])
            self.assertEqual(expected, set(files))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



# Copyright (C) 2024 Stewart Miles
# Licensed to the Python Software Foundation under a contributor agreement.
# See LICENSE.txt and CONTRIBUTORS.txt.

"""Python implementation of the calculate extension module."""

def fib(index):
    """Calculate a Fibonacci number.

    :param index: Index of the number in the Fibonacci sequence
      to calculate.

    :returns: Fibonacci number at the specified index.
      For example an index of 7 will return 13
    """
    current_value = 1
    previous_value = 0
    index -= 1
    while index > 0:
        next_value = current_value + previous_value
        previous_value = current_value
        current_value = next_value
        index -= 1
    return current_value
